from enum import Enum
from typing import TypeVar

import ida_hexrays
from ida_hexrays import mba_t, mblock_t, minsn_t, mop_t

from idahelper.microcode import mba, mblock


class TreeVisitOrder(Enum):
    PRE_ORDER = 0
    POST_ORDER = 1


class extended_microcode_visitor_t:
    def __init__(self, visit_order: TreeVisitOrder = TreeVisitOrder.PRE_ORDER):
        self.order: TreeVisitOrder = visit_order
        """How should we visit the tree"""
        self.mba: mba_t
        """The current function being visited"""
        self.blk: mblock_t
        """The current block being visited"""
        self.parents: list[mop_t | minsn_t]
        """List of parents for the current operand, which can be either a mop or a minsn"""
        self.top_ins: minsn_t
        """The top instruction"""
        self.prune: bool
        """Should skip sub-operands of the current operand? can be set from visit_X() methods"""

    def _visit_mop(self, op: mop_t) -> int:
        """Visit a mop, return 0 to continue visiting, any other value to stop visiting"""
        return 0

    def _visit_insn(self, ins: minsn_t) -> int:
        """Visit a minsn, return 0 to continue visiting, any other value to stop visiting"""
        return 0

    def __visit_minsn(self, ins: minsn_t) -> int:
        if self.order == TreeVisitOrder.PRE_ORDER:
            state, diff = 0, 1
        else:
            state, diff = 1, -1

        while state in [0, 1]:
            if state == 0:
                # Visit the instruction
                res = assert_not_none(self._visit_insn(ins))
                # If we should stop visit children (stop visiting or prune), return immediately
                if res != 0 or self.prune:
                    self.prune = False
                    return res
            elif state == 1:
                # Visit the instruction operands
                self.parents.append(ins)
                for op in [ins.l, ins.r, ins.d]:
                    if op is None:
                        continue

                    res = self.__visit_mop(op)
                    if res != 0:
                        return res
                self.parents.pop()

            # Update state
            state += diff

        return 0

    def __visit_mop(self, op: mop_t) -> int:  # noqa: C901
        if self.order == TreeVisitOrder.PRE_ORDER:
            state, diff = 0, 1
        else:
            state, diff = 1, -1

        if op.t == 0:
            # Invalid mop, skip.
            # It usually happens where there is a minsn embedded inside a mop.
            # The destination mop will exist in this case, but would be invalid.
            return 0

        while state in [0, 1]:
            if state == 0:
                # Visit the mop
                res = assert_not_none(self._visit_mop(op))
                if res != 0 or self.prune:
                    self.prune = False
                    return res
            elif state == 1:
                # Visit children
                minsn: minsn_t | None = None
                mops: list[mop_t] = []

                if op.t == ida_hexrays.mop_d:
                    minsn = op.d
                elif op.t == ida_hexrays.mop_f:
                    mops.extend(op.f.args)
                elif op.t == ida_hexrays.mop_a:
                    mops.append(op.a)
                elif op.t == ida_hexrays.mop_p:
                    mops.append(op.pair.lop)
                    mops.append(op.pair.hop)

                # Go over all children
                self.parents.append(op)

                for mop in mops:
                    if mop is None:
                        continue
                    res = self.__visit_mop(mop)
                    if res != 0:
                        return res

                if minsn is not None:
                    res = self.__visit_minsn(minsn)
                    if res != 0:
                        return res

                self.parents.pop()

            # Update state
            state += diff

        return 0

    def _init_state(self, blk: mblock_t):
        """Init the internal state of the visitor"""
        self.parents: list[mop_t | minsn_t] = []
        self.prune = False
        self.blk: mblock_t = blk
        self.mba: mba_t = blk.mba

    def _parent(self) -> minsn_t | mop_t:
        """Return the parent of the current operand"""
        if not self.parents:
            raise ValueError("No parent available")
        return self.parents[-1]

    def visit_instruction(self, blk: mblock_t, insn: minsn_t) -> int:
        """Run the visitor over a single instruction"""
        self._init_state(blk)
        self.top_ins = insn
        return self.__visit_minsn(insn)

    def visit_block(self, blk: mblock_t) -> int:
        """Run the visitor over the block"""
        self._init_state(blk)
        for insn in mblock.instructions(blk):
            self.top_ins = insn
            res = self.__visit_minsn(insn)
            if res != 0:
                return res
        return 0

    def visit_function(self, func_mba: mba_t) -> int:
        """Run the visitor over a function"""
        for blk in mba.blocks(func_mba):
            self._init_state(blk)
            res = self.visit_block(blk)
            if res != 0:
                return res
        return 0


T = TypeVar("T")


def assert_not_none(t: T | None) -> T:
    if t is None:
        raise ValueError("Accidentally passed none")
    return t



from collections.abc import Iterator

import ida_hexrays
import idaapi
from ida_hexrays import lvar_t, mba_ranges_t, mba_t, mblock_t


def blocks(mba: mba_t) -> Iterator[mblock_t]:
    """Create a generator of the block's instructions"""
    for i in range(mba.qty):
        yield mba.get_mblock(i)


def from_func(func_ea: int, opt_level: int = ida_hexrays.MMAT_LVARS) -> mba_t | None:
    """Get mba for a function for the given optimization level"""
    f = idaapi.get_func(func_ea)
    if f is None:
        return None
    return ida_hexrays.gen_microcode(mba_ranges_t(f), None, None, 0, opt_level)


def get_ret_lvar(mba: mba_t) -> lvar_t | None:
    """Get the return variable of the function"""
    if mba.retvaridx != -1:
        return mba.vars[mba.retvaridx]

    return None



import ida_hexrays
from ida_hexrays import (
    minsn_t,
)

from . import mop


def get_func_name_of_call(insn: minsn_t) -> str | None:
    """Given a call instruction, return the name of the called function"""
    assert insn.opcode == ida_hexrays.m_call, "Not a call instruction"
    return mop.get_name(insn.l)



__all__ = ["PacClient", "VtableXrefTuple"]

from .pac_interface import PacClient, VtableXrefTuple
from .pacxplorer import PacXplorerClient
from .pacxplorerng import PacXplorerNGPacClient


class StubClient(PacClient):
    def __init__(self):
        self.__client: PacClient | None = None

    def _client(self) -> PacClient:
        if self.__client is not None:
            return self.__client

        if PacXplorerNGPacClient.is_pac_plugin_installed():
            self.__client = PacXplorerNGPacClient()
        elif PacXplorerClient.is_pac_plugin_installed():
            self.__client = PacXplorerClient()
        else:
            PacXplorerNGPacClient.ensure_pac_plugin_installed()
        return self.__client

    def _pac_candidates_from_movk(self, movk_ea: int) -> list[VtableXrefTuple]:
        return self._client()._pac_candidates_from_movk(movk_ea)

    def pac_xrefs_to_func(self, func_ea: int) -> list[int]:
        return self._client().pac_xrefs_to_func(func_ea)

    @classmethod
    def is_pac_plugin_installed(cls) -> bool:
        return PacXplorerNGPacClient.is_pac_plugin_installed() or PacXplorerClient.is_pac_plugin_installed()


client: PacClient = StubClient()



import re
from collections.abc import Iterator
from functools import cache

import ida_bytes
import ida_nalt
import idaapi
import idautils
import idc

PTR_SIZE = 8
RETRY_COUNT = 20


def str_from_ea(ea: int) -> str | None:
    """Given EA return as string the C-String stored at that location"""
    content = idc.get_strlit_contents(ea)
    if content is None:
        return None
    return content.decode()


def name_from_ea(ea: int) -> str | None:
    """Given EA return the name of the symbol"""
    return idc.get_name(ea)


def qword_from_ea(ea: int) -> int:
    """Given EA return the 8 byte value stored at that location"""
    return ida_bytes.get_qword(ea)


def ea_from_name(name: str) -> int | None:
    """Given a name return the EA of the symbol"""
    ea = idc.get_name_ea(idaapi.BADADDR, name)
    if ea == idaapi.BADADDR:
        return None
    return ea


def set_name(ea: int, name: str, retry: bool = False, retry_count: int = RETRY_COUNT) -> bool:
    """Set the name of the symbol at EA to the given name"""
    res = bool(idc.set_name(ea, name, idc.SN_NOWARN))
    if res or not retry:
        return res

    cur_name = name_from_ea(ea)
    if cur_name is not None and re.match(re.escape(name) + r"(_\d+)?$", cur_name):
        # If the current name already has a postfix, we assume it was set by a previous retry
        return True

    for i in range(1, retry_count + 1):
        new_name = f"{name}_{i}"
        res = bool(idc.set_name(ea, new_name, idc.SN_NOWARN))
        if res:
            return res

    return False


def is_user_defined_name(ea: int) -> bool:
    """Check if the name at EA is user defined"""
    ea_flags = ida_bytes.get_full_flags(ea)
    if ea_flags == 0:
        print("[Error] EA is not valid")
        return False

    return ida_bytes.has_user_name(ea_flags)


def names() -> Iterator[tuple[int, str]]:
    """Return all the names in the binary"""
    return idautils.Names()


@cache
def imports() -> dict[int, str]:
    """Return a dictionary of all imports in the binary, mapping EA to import name."""
    all_imports = {}
    for i in range(ida_nalt.get_import_module_qty()):

        def import_callback(ea: int, name: str, _ordinal: int):
            all_imports[ea] = name
            return True

        ida_nalt.enum_import_names(i, import_callback)
    return all_imports


def name_from_imported_ea(ea: int) -> str | None:
    """If the ea is an import, return its name."""
    return imports().get(ea)



import functools

import ida_idp
from ida_hexrays import reg2mreg


@functools.cache
def cs_reg() -> int:
    """Returns the mreg id of the CS register"""
    return reg2mreg(ida_idp.ph_get_regnames().index("CS"))



import functools
import sys

from idautils import Strings

CACHED_STRINGS_ATTR = "_ida_strings"


@functools.cache
def strings():
    # Cache it somewhere else, to avoid build strings every time we reload our plugin
    main_module = sys.modules["__main__"]
    if hasattr(main_module, CACHED_STRINGS_ATTR):
        return getattr(main_module, CACHED_STRINGS_ATTR)

    strings_obj = Strings()
    setattr(main_module, CACHED_STRINGS_ATTR, strings_obj)
    return strings_obj


def find_str(content: str) -> Strings.StringItem:
    try:
        return next(s for s in strings() if str(s) == content)
    except StopIteration:
        raise ValueError(f"Could not find string {content}")  # noqa: B904


def find_strs(content: str) -> list[Strings.StringItem]:
    """find all the strings with the given content"""
    return [s for s in strings() if str(s) == content]



from collections.abc import Iterator

import ida_funcs
import ida_ida
import idaapi
from ida_funcs import func_t

FLAG_NO_RETURN = ida_funcs.FUNC_NORET
FLAG_OUTLINE = ida_funcs.FUNC_OUTLINE


def iterate_functions(start: int | None = None, end: int | None = None) -> Iterator[func_t]:
    """Iterate all the function in the given range or the whole project"""
    # Copied from idautils.functions
    if start is None:
        start = ida_ida.inf_get_min_ea()
    if end is None:
        end = ida_ida.inf_get_max_ea()

    chunk = ida_funcs.get_fchunk(start)
    if not chunk:
        chunk = ida_funcs.get_next_fchunk(start)
    while chunk and chunk.start_ea < end and (chunk.flags & ida_funcs.FUNC_TAIL) != 0:
        chunk = ida_funcs.get_next_fchunk(chunk.start_ea)
    func = chunk

    while func and func.start_ea < end:
        start_ea = func.start_ea
        yield func
        func = ida_funcs.get_next_func(start_ea)


def get_start_of_function(ea: int) -> int | None:
    """Get the beginning of the function the given address is in."""
    func = idaapi.get_func(ea)
    if func is None:
        return None
    return func.start_ea


def is_in_function(ea: int) -> bool:
    """Check if the given address is in a function."""
    return idaapi.get_func(ea) is not None


def add_function(start_ea: int, end_ea: int) -> bool:
    """Add a function with the given start and end addresses and name."""
    return idaapi.add_func(start_ea, end_ea)


def has_flags(func: int | func_t, flag: int) -> bool:
    """Check if function already has the given flag"""
    _func = idaapi.get_func(func) if isinstance(func, int) else func

    if _func is None:
        return False

    return _func.flags & flag == flag


def apply_flag_to_function(func: int | func_t, flag: int) -> bool:
    """Apply a flag to the function at the given address."""
    _func = idaapi.get_func(func) if isinstance(func, int) else func
    if _func is None:
        return False

    _func.flags |= flag
    return ida_funcs.update_func(_func)


def remove_flag_to_function(func: int | func_t, flag: int) -> bool:
    """Remove a flag from a function at the given address."""
    _func = idaapi.get_func(func) if isinstance(func, int) else func
    if _func is None:
        return False

    _func.flags &= ~flag
    return ida_funcs.update_func(_func)


def get_next_function(func: int | func_t) -> func_t | None:
    """Given a function, return the next function in the binary"""
    ea = func.start_ea if isinstance(func, func_t) else func
    return ida_funcs.get_next_func(ea)



import dataclasses
from typing import TypeVar

import ida_hexrays
import ida_nalt
import ida_typeinf
import idaapi
from ida_funcs import func_t
from ida_typeinf import func_type_data_t, tinfo_t, udm_t, udt_type_data_t

from idahelper.xrefs import get_xrefs_to

BOOL = tinfo_t(ida_typeinf.BT_BOOL)
VOID = tinfo_t(ida_typeinf.BT_VOID)


def from_c_type(c_type: str) -> tinfo_t | None:
    """Given a C type string, return matching `tinfo_t`"""
    tif = tinfo_t()
    if c_type == "void":
        tif.create_simple_type(ida_typeinf.BT_VOID)
        return VOID
    else:
        # noinspection PyTypeChecker
        if (
            ida_typeinf.parse_decl(
                tif,
                None,
                c_type + ";",
                ida_typeinf.PT_SIL | ida_typeinf.PT_NDC | ida_typeinf.PT_TYP,
            )
            is not None
        ):
            return tif
    return None


def from_size(size: int) -> tinfo_t | None:
    """Convert number of bytes to `tinfo_t`"""
    # Using those types seems to make IDA hide casts
    if size == 1:
        return tinfo_t(ida_typeinf.BT_UNK_BYTE)
    elif size == 2:
        return tinfo_t(ida_typeinf.BT_UNK_WORD)
    elif size == 4:
        return tinfo_t(ida_typeinf.BT_UNK_DWORD)
    elif size == 8:
        return tinfo_t(ida_typeinf.BT_UNK_QWORD)
    else:
        print(f"[Error] unsupported size {size}")
        return None


def from_struct_name(name: str) -> tinfo_t | None:
    """Given a struct name, return matching `tinfo_t`"""
    tif = tinfo_t()
    if not tif.get_named_type(ida_typeinf.get_idati(), name, ida_typeinf.BTF_TYPEDEF, True, False):
        return None
    return tif


def from_func(func: func_t) -> tinfo_t | None:
    """Given a function, return matching `tinfo_t`"""
    return from_ea(func.start_ea)


def from_ea(ea: int) -> tinfo_t | None:
    """Given ea, return matching `tinfo_t`"""
    tif = tinfo_t()
    if idaapi.get_tinfo(tif, ea):
        return tif
    return None


def get_func_details(func: func_t | tinfo_t) -> func_type_data_t | None:
    """Given a function, return its type details"""
    # Convert to tif
    if isinstance(func, func_t):
        func = from_func(func)
        if func is None:
            return None

    func_type = func_type_data_t()
    if func.get_func_details(func_type):
        return func_type


def from_func_details(details: func_type_data_t) -> tinfo_t | None:
    """Given a function type details, return matching `tinfo_t`"""
    tif = tinfo_t()
    if tif.create_func(details):
        return tif


def apply_tinfo_to_func(tif: tinfo_t, func: func_t) -> bool:
    """Apply typing info to the given function`"""
    return apply_tinfo_to_ea(tif, func.start_ea)


def apply_tinfo_to_ea(tif: tinfo_t, ea: int) -> bool:
    """Apply typing info to the given ea`"""
    return idaapi.apply_tinfo(ea, tif, idaapi.TINFO_DEFINITE)


def apply_tinfo_to_call(tif: tinfo_t, call_ea: int) -> bool:
    """Apply typing info to the given call instruction"""
    return ida_nalt.set_op_tinfo(call_ea, 0, tif)


@dataclasses.dataclass
class FuncParam:
    type: str
    name: str | None = None


def from_func_components(return_type: str, parameters: list[FuncParam]) -> tinfo_t | None:
    """Create a tif from return type and list of parameters"""
    params_str = ",".join(f"{p.type} {p.name or ''}" for p in parameters)
    sig = f"{return_type} f({params_str})"
    return from_c_type(sig)


def pointer_of(tif: tinfo_t) -> tinfo_t:
    """Given a tif, return tif of pointer to the type"""
    return ida_hexrays.make_pointer(tif)


def get_udt(tif: tinfo_t) -> udt_type_data_t | None:
    """Get udt from tif"""
    if not tif.is_struct():
        print(f"[Error] get_udt: Not a struct type - {tif.dstr()}")
        return None

    udt_data = udt_type_data_t()
    if not tif.get_udt_details(udt_data):
        return None
    return udt_data


def get_member(tif: tinfo_t, offset: int) -> udm_t | None:
    """Get member of a struct at given offset"""
    if not tif.is_struct():
        print(f"[Error] get_member: Not a struct type - {tif.dstr()}")
        return None

    udm = udm_t()
    udm.offset = offset * 8
    if tif.find_udm(udm, ida_typeinf.STRMEM_OFFSET) == -1:
        return None

    return udm


def get_member_recursive(tif: tinfo_t, offset: int) -> tuple[tinfo_t, udm_t] | None:
    """Get member of a struct at given offset. If it comes from the base class, return it with the base class type."""
    t: tinfo_t | None = tif
    while t is not None:
        udm = get_member(t, offset)
        if udm is None:
            return None
        elif not udm.is_baseclass():
            return t, udm

        t = get_parent_class(t)
    return None


def get_parent_classes(tif: tinfo_t, including_current_type: bool = False) -> list[tinfo_t]:
    """Get parent classes of a struct. For example: IOService -> [IORegistryEntry, OSObject]"""
    classes: list[tinfo_t] = []
    if including_current_type:
        classes.append(tif)

    while (parent := get_parent_class(tif)) is not None:
        classes.append(parent)
        tif = parent

    return classes


def get_parent_class(tif: tinfo_t) -> tinfo_t | None:
    """Get parent class of a struct"""
    if not tif.is_struct():
        return None

    udt_data = get_udt(tif)
    if udt_data is None:
        return None

    if not udt_data.is_cppobj():
        return None

    for udm in udt_data:
        if udm.is_baseclass():
            # Copy the type as it somehow got freed...
            parent_type: tinfo_t = tinfo_t(udm.type)
            return parent_type


def get_base_offset_for_class(tif: tinfo_t) -> int | None:
    """Given a cpp class, return the offset in bytes of its first own member"""
    if not tif.is_struct():
        return None

    udt_data = get_udt(tif)
    if udt_data is None:
        return None

    if not udt_data.is_cppobj():
        return 0

    for udm in udt_data:
        udm: udm_t
        if udm.is_baseclass():
            continue

        # Convert from bits to bytes
        return int((udm.offset + 7) / 8)

    # No members
    return None


def get_children_classes(tif: tinfo_t) -> list[tinfo_t] | None:
    """Get child classes of a struct."""
    if not tif.is_struct():
        return None

    tid = tif.get_tid()
    children = []
    for xref in get_xrefs_to(tid, is_data=True):
        name = ida_typeinf.get_tid_name(xref)
        if name and ".baseclass_" in name:
            cls_name = name.split(".baseclass_")[0]
            cls = from_struct_name(cls_name)
            if cls is not None:
                children.append(cls)
                children.extend(get_children_classes(cls) or [])
    return children


def vtable_type_from_type(tif: tinfo_t) -> tinfo_t | None:
    """Get vtable for a class"""
    if not tif.is_struct():
        return None

    if not tif.has_vftable():
        return None

    # noinspection PyTypeChecker
    return from_struct_name(tif.get_type_name() + "_vtbl")


def type_from_vtable_type(tif: tinfo_t) -> tinfo_t | None:
    """Get class type from vtable type"""
    # noinspection PyUnresolvedReferences
    if not tif.is_vftable() or not tif.get_type_name().endswith("_vtbl"):
        return None

    # noinspection PyTypeChecker
    return from_struct_name(tif.get_type_name()[:-5])


def set_udm_type(tif: tinfo_t, udm: udm_t, udm_type: tinfo_t) -> bool:
    """For a `udm` of a `tif`, set its type"""
    index = get_udm_index(tif, udm)
    if index is None:
        return False

    return tif.set_udm_type(index, udm_type) == 0


def set_udm_name(tif: tinfo_t, udm: udm_t, new_name: str) -> bool:
    """For a `udm` of a `tif`, set its name"""
    index = get_udm_index(tif, udm)
    if index is None:
        return False

    return tif.rename_udm(index, new_name) == 0


def tid_from_udm(tif: tinfo_t, udm: udm_t) -> int | None:
    """Get the type ID of a `udm` in a `tif`"""
    index = get_udm_index(tif, udm)
    if index is None:
        return None
    return tif.get_udm_tid(index)


def get_udm_index(tif: tinfo_t, udm: udm_t) -> int | None:
    """Get the index of a `udm` in a `tif`"""
    index = tif.find_udm(udm, ida_typeinf.STRMEM_OFFSET)
    if index == -1:
        return None
    return index


def create_from_c_decl(decl: str) -> bool:
    """Create a new type definition from `decl`"""
    return not ida_typeinf.idc_parse_types(decl, 0)


def get_common_ancestor(types: list[tinfo_t]) -> tinfo_t | None:
    """Given list of C++ types, return their common ancestor"""
    if not types:
        return None
    if len(types) == 1:
        return types[0]

    current_common: tinfo_t = types[0]
    current_common_parents: list[tinfo_t] = get_parent_classes(current_common, True) or []
    for typ in types:
        if typ == current_common:
            continue
        typ_parents = get_parent_classes(typ, True) or []
        ancestor_res = _find_common_ancestor(current_common_parents, typ_parents)
        if ancestor_res is None:
            print(f"[Error] could not find common ancestor between {current_common.dstr()} and {typ.dstr()}")
            return None
        new_common, new_common_parents = ancestor_res
        if new_common != current_common:
            current_common, current_common_parents = new_common, new_common_parents

    return current_common


T = TypeVar("T")


def _find_common_ancestor(path1: list[T], path2: list[T]) -> tuple[T, list[T]] | None:
    """
    Given two reversed paths (from node to root), return the lowest common ancestor.

    Example:
        path1 = ['D', 'B', 'A']
        path2 = ['E', 'B', 'A']
        -> returns 'B', ['B', 'A']
    """
    i1 = len(path1) - 1
    i2 = len(path2) - 1
    common: T | None = None

    while i1 >= 0 and i2 >= 0 and path1[i1] == path2[i2]:
        common = path1[i1]
        i1 -= 1
        i2 -= 1

    if common is None:
        return None

    return common, path1[i1 + 1 :]



from asyncio import Protocol
from collections.abc import Callable

from ida_hexrays import mblock_t, minsn_t, optblock_t, optinsn_t


class CounterMixin:
    cnt: int = 0

    def count(self, amount: int = 1):
        self.cnt += amount


class optblock_counter_t(optblock_t, CounterMixin, Protocol): ...


class optinsn_counter_t(optinsn_t, CounterMixin, Protocol): ...


class optblock_counter_wrapper_t(optblock_t):
    def __init__(self, factory: Callable[[], optblock_counter_t]):
        super().__init__()
        self.factory = factory

    def func(self, blk: mblock_t) -> int:
        optimizer = self.factory()
        optimizer.func(blk)
        return optimizer.cnt


class optinsn_counter_wrapper_t(optinsn_t):
    def __init__(self, factory: Callable[[], optinsn_counter_t]):
        super().__init__()
        self.factory = factory

    def func(self, blk: mblock_t, ins: minsn_t, optflags: int) -> int:
        optimizer = self.factory()
        optimizer.func(blk, ins, optflags)
        return optimizer.cnt



import ida_hexrays
from ida_hexrays import lvar_ref_t, lvar_t, mba_t, mop_addr_t, mop_t

from idahelper import memory
from idahelper.ast import lvars


def from_global_ref(ea: int) -> mop_t:
    """Given `ea` of a global address, create a mop that represents a `(void*)ea`"""
    mop = mop_t()
    mop.t = ida_hexrays.mop_a
    mop.a = mop_addr_t()
    mop.a.t = ida_hexrays.mop_v
    mop.a.g = ea
    mop.size = 8
    return mop


def from_lvar(lvar: lvar_t, mba: mba_t, offset: int = 0) -> mop_t:
    """Given a lvar, create a mop that represents it"""
    mop = mop_t()
    mop.t = ida_hexrays.mop_l
    mop.l = lvar_ref_t(mba, lvars.get_index(mba.vars, lvar), offset)
    mop.size = lvar.width
    return mop


def from_const_value(value: int, size: int) -> mop_t:
    """Given a const value, create a mop that represents it"""
    mop = mop_t()
    mop.make_number(value, size)
    return mop


def get_name(mop: mop_t) -> str | None:
    """Given a mop representing a symbol/helper, return its name"""
    if mop.helper is not None:
        return mop.helper
    elif mop.g is not None:
        return memory.name_from_ea(mop.g)


def get_str(mop: mop_t) -> str | None:
    """Given a mop representing a string, return its value"""
    if mop.t == ida_hexrays.mop_str:
        return mop.cstr
    elif mop.is_glbaddr():
        return memory.str_from_ea(mop.a.g) or None


def get_const_int(mop: mop_t, is_signed: bool = False) -> int | None:
    """Given a mop representing a const int, return its value"""
    if mop.t == ida_hexrays.mop_n:
        if is_signed:
            return mop.signed_value()
        else:
            return mop.unsigned_value()


def get_local_variable(mop: mop_t) -> lvar_t | None:
    """Given a mop representing a local variable, return the variable"""
    if mop.t == ida_hexrays.mop_l:
        return mop.l.var()


def get_stack_offset(mop: mop_t) -> int | None:
    """Given a mop representing a stack address, return its offset"""
    offset = None
    if mop.t == ida_hexrays.mop_l:
        offset = mop.l.var().get_stkoff() + mop.l.off
    elif mop.t == ida_hexrays.mop_S:
        offset = mop.s.off
    elif mop.t == ida_hexrays.mop_a:
        offset = get_stack_offset(mop.a)
    return offset if offset is not None and offset != -1 else None



from typing import NamedTuple, Protocol

from ..pac_interface import VtableXrefTuple


class PacTuple(NamedTuple):
    """
    ```
    LDR     X8, [X0]            ; load vtable address
    LDRA    X9, [X8,#0x18]!     ; X8 = vtbl + offset
    MOVK    X8, #0x68DA,LSL#48  ; set the hash
    BLRAA   X9, X8              ; virtual call
    ```

    This tuple represents the PAC (Pointer Authentication Code) information
    extracted from the MOVK instruction.
    It contains the offset within the vtable and the PAC value.
    The offset is the immediate value used in the MOVK instruction,
    """

    offset: int
    pac: int


class MovkCodeTuple(NamedTuple):
    """This tuple represents the MOVK instruction and its associated PAC information."""

    pac_tuple: PacTuple
    movk_addr: int


class ExplorerProtocol(Protocol):
    def analyze(self, only_cached: bool = False) -> None:
        """
        Analyze the current IDA database.

        If only_cached is True, only cached values will be used. An exception will be raised if no cached values are available.
        If not, the analysis will be performed and results will be cached.
        """

    def movk_to_functions(self, movk_ea: int) -> list[VtableXrefTuple]:
        """Get the possible function candidates for a given MOVK instruction address."""

    def function_to_movks(self, func_ea: int) -> list[int]:
        """Get the possible MOVK candidates for a given function."""

    def is_pac_function(self, func_ea: int) -> bool:
        """Check if the function at the given address is a PAC function."""

    @staticmethod
    def get_instance() -> "ExplorerProtocol":
        """Get the singleton instance of the Explorer."""


def get_explorer_instance() -> ExplorerProtocol:
    """Get the singleton instance of the Explorer."""
    # noinspection PyUnresolvedReferences
    from pacxplorerng.explorer import Explorer

    return Explorer.get_instance()



from collections.abc import Callable

import ida_hexrays
import idaapi
from ida_funcs import func_t
from ida_hexrays import CV_FAST, cexpr_t, cfunc_t, cfuncptr_t, cinsn_t, ctree_visitor_t, lvar_t, lvars_t


def from_ea(ea: int, should_generate_pseudocode: bool = False) -> cfuncptr_t | None:
    """decompile a function"""
    f = idaapi.get_func(ea)
    if f is None:
        return None

    return from_func(f, should_generate_pseudocode)


def from_func(func: func_t, should_generate_pseudocode: bool = False) -> cfuncptr_t | None:
    """Decompile a function"""
    decompiled = idaapi.decompile(func)
    if decompiled is None:
        return None

    if should_generate_pseudocode:
        decompiled.get_pseudocode()

    return decompiled


def get_lvar_by_offset(func: cfunc_t | cfuncptr_t, offset: int) -> lvar_t | None:
    """Given a decompiled function, return the lvar located at {offset} on the stack"""
    lvars: lvars_t = func.get_lvars()
    for lv in lvars:
        if lv.get_stkoff() == offset:
            return lv
    return None


def get_arg_index(func: cfunc_t | cfuncptr_t, var: lvar_t) -> int | None:
    """Given a decompiled function, return the argument index of the given {var}"""
    lvars: lvars_t = func.get_lvars()
    for var_index, lv in enumerate(lvars):
        if lv.name == var.name:
            for arg_index, index_in_vars in enumerate(func.argidx):
                if index_in_vars == var_index:
                    return arg_index
    return None


class _ExpressionFinderVisitor(ctree_visitor_t):
    """Search for expressions fulfilling a condition in ctree"""

    def __init__(self, match_fn: Callable[[cexpr_t], bool]):
        super().__init__(CV_FAST)
        self.match_fn: Callable[[cexpr_t], bool] = match_fn
        self.found: list[cexpr_t] = []

    def visit_expr(self, expr: cexpr_t) -> int:  # pyright: ignore[reportIncompatibleMethodOverride]
        if self.match_fn(expr):
            self.found.append(expr)
        return 0


class _StatementFinderVisitor(ctree_visitor_t):
    """Search for expressions fulfilling a condition in ctree"""

    def __init__(self, match_fn: Callable[[cinsn_t], bool]):
        super().__init__(CV_FAST)
        self.match_fn: Callable[[cinsn_t], bool] = match_fn
        self.found: list[cinsn_t] = []

    def visit_insn(self, insn: cinsn_t) -> int:  # pyright: ignore[reportIncompatibleMethodOverride]
        if self.match_fn(insn):
            self.found.append(insn)
        return 0


def get_call_expression_at_ea(func: cfunc_t | cfuncptr_t, call_ea: int) -> cexpr_t | None:
    """Given a function and ea, find the call expression that are on this ea, or none if not found"""
    # Note: there might be multiple results if the ea is inside an if condition (as the "if" statement will be returned,
    # and apply to will apply to the body as well), so we try to narrow it down by checking the ea in the condition.
    finder = _ExpressionFinderVisitor(lambda e: e.op == ida_hexrays.cot_call and e.ea == call_ea)
    for insn in func.get_eamap().get(call_ea, []):
        finder.apply_to(insn, None)  # pyright: ignore[reportArgumentType]
    if len(finder.found) != 1:
        return None
    return finder.found[0]


def get_return_statements(func: cfunc_t | cfuncptr_t) -> list[cexpr_t]:
    """Given a function, return all the expressions that were returned from the function"""
    finder = _StatementFinderVisitor(lambda e: e.op == ida_hexrays.cit_return)
    finder.apply_to(func.body, None)  # pyright: ignore[reportArgumentType]

    return [insn.creturn.expr for insn in finder.found]



__all__ = ["PacClient", "VtableXrefTuple"]

from abc import ABC, abstractmethod
from typing import NamedTuple

import idaapi
from ida_loader import find_plugin
from ida_typeinf import tinfo_t

from idahelper import cpp, instructions, memory

MAX_PREVIOUS_OPCODES_FOR_MOVK_SCAN = 10
MAX_NEXT_OPCODES_FOR_BLR_SCAN = MAX_PREVIOUS_OPCODES_FOR_MOVK_SCAN


class VtableXrefTuple(NamedTuple):
    xref_to: int
    vtable_addr: int
    vtable_entry_addr: int
    offset: int
    pac: int


class PacClient(ABC):
    """Client for PacExplorer plugin for querying PAC xrefs."""

    plugin_name: str
    """Name of the plugin to use for PAC xrefs"""

    @classmethod
    def is_pac_plugin_installed(cls) -> bool:
        """Check if the PAC plugin is installed"""
        return find_plugin(cls.plugin_name) is not None

    @classmethod
    def ensure_pac_plugin_installed(cls):
        """Raise an exception if the PAC plugin is not installed"""
        if not cls.is_pac_plugin_installed():
            raise AssertionError("PacXplorer or PacXplorerNG plugin are not installed. Please install it...")

    @abstractmethod
    def _pac_candidates_from_movk(self, movk_ea: int) -> list[VtableXrefTuple]:
        """Given the EA of a movk, return possible functions that could be called using this movk"""

    def pac_candidates_from_movk(self, movk_ea: int) -> list[int]:
        """Given the EA of a movk, return possible functions that could be called using this movk"""
        return [item.xref_to for item in self._pac_candidates_from_movk(movk_ea)]

    def pac_candidates_for_call(self, call_ea: int) -> list[int]:
        """Given the EA of a call, return possible functions that could be called from this call"""
        movk_ea = self.get_previous_movk(call_ea)
        if movk_ea is None:
            return []
        return self.pac_candidates_from_movk(movk_ea)

    def pac_class_candidates_from_movk(self, movk_ea: int) -> list[tinfo_t]:
        """Given the EA of a movk, return the classes that implements the function that could be called using this movk"""
        candidates = self._pac_candidates_from_movk(movk_ea)
        if candidates is None:
            return []

        types: list[tinfo_t] = []
        for candidate in candidates:
            vtable_addr = candidate.vtable_addr
            vtable_name = memory.name_from_ea(vtable_addr)
            if vtable_name is None:
                print(f"[Error] vtable name is none at {vtable_addr:X}, aborting PAC solver.")
                return []
            original_type = cpp.type_from_vtable_name(vtable_name)
            if original_type is None:
                print(f"[Error] failed to convert vtable to type. Vtable at {vtable_addr:X}, name: {vtable_name}")
                return []
            types.append(original_type)
        return types

    @abstractmethod
    def pac_xrefs_to_func(self, func_ea: int) -> list[int]:
        """Get all xrefs to the given PAC function EA"""

    def pac_calls_xrefs_to_func(self, func_ea: int) -> list[int]:
        """Get the call instructions that use the PAC xrefs to the given function EA"""
        movks = self.pac_xrefs_to_func(func_ea)
        calls = []
        for movk in movks:
            call = self.get_next_blr(movk)
            if call is not None:
                calls.append(call)
        return calls

    @staticmethod
    def get_previous_movk(call_ea: int) -> int | None:
        """Given a call, search previous instructions to find a movk call"""
        insn = instructions.decode_instruction(call_ea)
        if not insn:
            return None

        if insn.get_canon_mnem() not in ("BLR", "BR"):
            return None

        # Get the register for PAC code
        movk_reg = insn[1].reg
        # BLR with just one register is unauthenticated, so there will be no PAC xref
        if movk_reg == 0:
            return None

        for _ in range(MAX_PREVIOUS_OPCODES_FOR_MOVK_SCAN):
            insn = instructions.decode_previous_instruction(insn)
            # No more instructions in this execution flow
            if insn is None:
                break
            if insn.get_canon_mnem() == "MOVK" and insn[0].reg == movk_reg:
                return insn.ea
        return None

    @staticmethod
    def get_next_blr(movk_ea: int) -> int | None:
        """Given a movk, search next instructions to find a call"""
        insn = instructions.decode_instruction(movk_ea)
        if not insn:
            return None

        if insn.get_canon_mnem() != "MOVK":
            return None

        movk_reg = insn[0].reg
        func = idaapi.get_func(insn.ea)
        if func is None:
            return None

        for _ in range(MAX_NEXT_OPCODES_FOR_BLR_SCAN):
            insn = instructions.decode_next_instruction(insn, func)
            # No more instructions in this execution flow
            if insn is None:
                break
            if insn.get_canon_mnem() in ("BLR", "BR") and insn[1].reg == movk_reg:
                return insn.ea
        return None



from collections.abc import Iterator

from ida_hexrays import mblock_t, minsn_t


def instructions(blk: mblock_t) -> Iterator[minsn_t]:
    """Create a generator of the block's instructions"""
    ins: minsn_t = blk.head
    while ins is not None:
        yield ins
        ins = ins.next



import ida_bytes
import ida_hexrays
import idaapi
from ida_hexrays import carg_t, carglist_t, cexpr_t, cfunc_t, cfuncptr_t, lvar_t, number_format_t, var_ref_t
from ida_typeinf import tinfo_t

from idahelper import memory
from idahelper.ast import lvars


def get_call_name(call_expr: cexpr_t) -> str | None:
    """Get the name of the called function from a call expression."""
    assert call_expr.op == ida_hexrays.cot_call, "Expected a call expression"
    called_func: cexpr_t = call_expr.x
    if called_func.op == ida_hexrays.cot_helper:
        return called_func.helper
    elif called_func.op == ida_hexrays.cot_obj:
        return memory.name_from_ea(called_func.obj_ea)

    return None


def strip_casts(expr: cexpr_t) -> cexpr_t:
    """Strip casts from the expression."""
    while expr.op == ida_hexrays.cot_cast:
        expr = expr.x
    return expr


def from_cast(expr: cexpr_t, typ: tinfo_t) -> cexpr_t:
    """Create a cast expression from an existing expression and a type."""
    cast_expr = cexpr_t()
    cast_expr.op = ida_hexrays.cot_cast
    cast_expr.x = cexpr_t(expr)
    cast_expr.type = typ
    return cast_expr


def from_var(var: var_ref_t) -> cexpr_t:
    """Create a cexpr_t from a var_ref_t."""
    var_t: lvar_t = var.getv()
    var_expr = cexpr_t()
    var_expr.op = ida_hexrays.cot_var
    var_expr.v = var
    var_expr.type = var_t.type()
    return var_expr


def from_lvar_index(lvar_index: int, func: cfuncptr_t) -> cexpr_t:
    """Create a cexpr_t from a lvar index in func."""
    var_ref = var_ref_t()
    var_ref.mba = func.mba
    var_ref.idx = lvar_index

    return from_var(var_ref)


def from_var_name(name: str, func: cfuncptr_t) -> cexpr_t:
    """Create a cexpr_t from a variable name and the container function."""
    return from_lvar_index(lvars.get_index_by_name(func.get_lvars(), name), func)


def from_const_value(
    x: int, cur_func: cfuncptr_t | cfunc_t | None = None, ea: int = idaapi.BADADDR, is_hex: bool = False
) -> cexpr_t:
    """Create a cexpr_t from a constant value."""
    num = ida_hexrays.make_num(x, cur_func, ea)
    if is_hex:
        nf: number_format_t = num.n.nf
        nf.flags = ida_bytes.hex_flag()
        nf.flags32 = nf.flags & 0xFF_FF_FF_FF
    return num


def from_binary_op(lhs: cexpr_t, rhs: cexpr_t, op: int, typ: tinfo_t, ea: int = idaapi.BADADDR) -> cexpr_t:
    """Create the binary operation expression `lhs op rhs`."""
    bin_expr = cexpr_t()
    bin_expr.ea = ea
    bin_expr.type = typ
    bin_expr.op = op
    bin_expr.x = lhs
    bin_expr.y = rhs
    return bin_expr


def from_assignment(lhs: cexpr_t, rhs: cexpr_t) -> cexpr_t:
    """Create an assignment expression from `lhs` and `rhs`."""
    assign_expr = cexpr_t()
    assign_expr.op = ida_hexrays.cot_asg
    assign_expr.x = lhs
    assign_expr.y = rhs
    assign_expr.type = lhs.type
    return assign_expr


def call_helper_from_sig(name: str, ret_type: tinfo_t, args: list[cexpr_t | carg_t]) -> cexpr_t:
    """Create a call expression from a name, arguments and return type"""
    return ida_hexrays.call_helper(ret_type, arglist_from_expr_arr(*args), name)


def from_call_and_args(call: cexpr_t, *args: cexpr_t | carg_t) -> cexpr_t:
    """Create a call expression from a call and arguments."""
    call_expr = cexpr_t()
    call_expr.op = ida_hexrays.cot_call
    call_expr.x = call
    call_expr.a = arglist_from_expr_arr(*args)
    return call_expr


def arglist_from_expr_arr(*args: cexpr_t | carg_t) -> carglist_t:
    """Convert a list of cexpr_t to a carglist_t."""
    arglist = carglist_t()
    for arg in args:
        if arg is None:
            print("[Warning]: argument is None")
            continue

        arglist.push_back(carg_from_expr(arg))
    return arglist


def carg_from_expr(expr: cexpr_t) -> carg_t:
    """Convert a cexpr_t to a carg_t."""
    if isinstance(expr, carg_t):
        return expr
    else:
        carg = carg_t()
        carg.assign(expr)
    return carg
